from abc import ABC, abstractmethod
from typing import Optional, List, Type

from sqlalchemy import Result
from pydantic import BaseModel, Field


class SQLExecuteResult(BaseModel):
    rowcount: int = Field(0)
    success: bool = Field(False)
    message: Optional[str] = Field(None)


class QueryResult(ABC):
    @abstractmethod
    def to_list(self) -> List[dict]:
        pass

    @abstractmethod
    def to_pydantic(self) -> List[BaseModel]:
        pass

    @abstractmethod
    def to_pandas(self):
        pass


class SQLModelQueryResult(QueryResult):
    def __init__(self, data: List[BaseModel]):
        self._data = data

    def to_list(self) -> List[dict]:
        return [item.model_dump() for item in self._data]

    def to_pydantic(self) -> List[BaseModel]:
        return self._data

    def to_pandas(self):
        try:
            import pandas as pd
        except Exception:
            raise ImportError(
                "Failed to import pandas, please install it with `pip install pandas`"
            )
        data = [item.model_dump() for item in self._data]
        return pd.DataFrame(data)


class SQLQueryResult(QueryResult):
    _result: Result

    def __init__(self, result):
        self._result = result

    def scalar(self):
        return self._result.scalar()

    def one(self):
        return self._result.one()

    def to_rows(self):
        return self._result.fetchall()

    def to_pandas(self):
        try:
            import pandas as pd
        except Exception:
            raise ImportError(
                "Failed to import pandas, please install it with `pip install pandas`"
            )
        keys = self._result.keys()
        rows = self._result.fetchall()
        return pd.DataFrame(rows, columns=keys)

    def to_list(self) -> List[dict]:
        keys = self._result.keys()
        rows = self._result.fetchall()
        return [dict(zip(keys, row)) for row in rows]

    def to_pydantic(self, model: Type[BaseModel]) -> List[BaseModel]:
        items = self.to_list()
        return [model.model_validate(item) for item in items]



import re
from typing import Any, Dict, List, Optional, Union

import sqlalchemy
from sqlalchemy import BinaryExpression, text

from pytidb.schema import TableModel


Filters = Union[Dict[str, Any], str, BinaryExpression]


# SQL filter operators:


def build_filter_clauses(
    filters: Filters, columns: Dict, table_model: TableModel
) -> List[BinaryExpression]:
    if isinstance(filters, dict):
        filter_clauses = build_dict_filter_clauses(filters, columns, table_model)
        return filter_clauses
    elif isinstance(filters, str):
        return [text(filters)]
    else:
        return [filters]


# Dict filter operators:

AND, OR, IN, NIN, GT, GTE, LT, LTE, EQ, NE = (
    "$and",
    "$or",
    "$in",
    "$nin",
    "$gt",
    "$gte",
    "$lt",
    "$lte",
    "$eq",
    "$ne",
)

COMPARE_OPERATOR = [IN, NIN, GT, GTE, LT, LTE, EQ, NE]

JSON_FIELD_PATTERN = re.compile(
    r"^(?P<column>[a-zA-Z_][a-zA-Z0-9_]*)\.(?P<json_field>[a-zA-Z_][a-zA-Z0-9_]*)$"
)


def build_dict_filter_clauses(
    filters: Dict[str, Any] | None, columns: Dict, table_model: TableModel
) -> List[BinaryExpression]:
    if filters is None:
        return []

    filter_clauses = []
    for key, value in filters.items():
        if key.lower() == AND:
            if not isinstance(value, list):
                raise TypeError(
                    f"Expect a list value for $and operator, but got {type(value)}"
                )
            and_clauses = []
            for item in value:
                and_clauses.extend(build_filter_clauses(item, columns, table_model))
            if len(and_clauses) == 0:
                continue
            filter_clauses.append(sqlalchemy.and_(*and_clauses))
        elif key.lower() == OR:
            if not isinstance(value, list):
                raise TypeError(
                    f"Expect a list value for $or operator, but got {type(value)}"
                )
            or_clauses = []
            for item in value:
                or_clauses.extend(build_filter_clauses(item, columns, table_model))
            if len(or_clauses) == 0:
                continue
            filter_clauses.append(sqlalchemy.or_(*or_clauses))
        elif key in columns:
            column = getattr(columns, key)
            if isinstance(value, dict):
                filter_clause = build_dict_column_filter(column, value)
            else:
                # implicit $eq operator: value maybe int / float / string
                filter_clause = build_dict_column_filter(column, {EQ: value})
            if filter_clause is not None:
                filter_clauses.append(filter_clause)
        elif "." in key:
            match = JSON_FIELD_PATTERN.match(key)
            if not match:
                raise ValueError(
                    f"Got unexpected filter key: {key}, please use valid column name instead"
                )
            column_name = match.group("column")
            json_field = match.group("json_field")
            column = sqlalchemy.func.json_extract(
                getattr(table_model, column_name), f"$.{json_field}"
            )
            if isinstance(value, dict):
                filter_clause = build_dict_column_filter(column, value)
            else:
                # implicit $eq operator: value maybe int / float / string
                filter_clause = build_dict_column_filter(column, {EQ: value})
            if filter_clause is not None:
                filter_clauses.append(filter_clause)
        else:
            raise ValueError(
                f"Got unexpected filter key: {key}, please use valid column name instead"
            )

    return filter_clauses


def build_dict_column_filter(
    column: Any, conditions: Dict[str, Any]
) -> Optional[BinaryExpression]:
    column_filters = []
    for operator, val in conditions.items():
        op = operator.lower()
        if op == IN:
            column_filters.append(column.in_(val))
        elif op == NIN:
            column_filters.append(~column.in_(val))
        elif op == GT:
            column_filters.append(column > val)
        elif op == GTE:
            column_filters.append(column >= val)
        elif op == LT:
            column_filters.append(column < val)
        elif op == LTE:
            column_filters.append(column <= val)
        elif op == NE:
            column_filters.append(column != val)
        elif op == EQ:
            column_filters.append(column == val)
        else:
            raise ValueError(
                f"Unknown filter operator {operator}. Consider using "
                "one of $in, $nin, $gt, $gte, $lt, $lte, $eq, $ne."
            )
    if len(column_filters) == 0:
        return None
    elif len(column_filters) == 1:
        return column_filters[0]
    else:
        return sqlalchemy.and_(*column_filters)



from contextlib import contextmanager
from contextvars import ContextVar
from typing import List, Literal, Optional, Type, Generator

from pydantic import PrivateAttr
import sqlalchemy
from sqlalchemy import (
    Executable,
    Inspector,
    SelectBase,
    text,
    Result,
)
from sqlalchemy.engine import Engine, create_engine
from sqlalchemy.orm import Session, DeclarativeMeta

from pytidb.base import Base, default_registry
from pytidb.databases import create_database, database_exists
from pytidb.schema import TableModel
from pytidb.table import Table
from pytidb.utils import build_tidb_dsn, create_engine_without_db
from pytidb.logger import logger
from pytidb.result import SQLExecuteResult, SQLQueryResult


SESSION = ContextVar[Session | None]("session", default=None)


class TiDBClient:
    _db_engine: Engine = PrivateAttr()
    _inspector: Inspector = PrivateAttr()

    def __init__(self, db_engine: Engine):
        self._db_engine = db_engine
        self._inspector = sqlalchemy.inspect(self._db_engine)
        self._identifier_preparer = self._db_engine.dialect.identifier_preparer

    @classmethod
    def connect(
        cls,
        database_url: Optional[str] = None,
        *,
        host: Optional[str] = "localhost",
        port: Optional[int] = 4000,
        username: Optional[str] = "root",
        password: Optional[str] = "",
        database: Optional[str] = "test",
        enable_ssl: Optional[bool] = None,
        ensure_db: Optional[bool] = False,
        debug: Optional[bool] = None,
        **kwargs,
    ) -> "TiDBClient":
        if database_url is None:
            database_url = str(
                build_tidb_dsn(
                    host=host,
                    port=port,
                    username=username,
                    password=password,
                    database=database,
                    enable_ssl=enable_ssl,
                )
            )

        if ensure_db:
            try:
                temp_engine = create_engine_without_db(
                    database_url, echo=debug, **kwargs
                )
                if not database_exists(temp_engine, database):
                    create_database(temp_engine, database)
            except Exception as e:
                logger.error(f"Failed to ensure database exists: {str(e)}")
                raise

        db_engine = create_engine(database_url, echo=debug, **kwargs)

        return cls(db_engine)

    def disconnect(self) -> None:
        self._db_engine.dispose()

    @property
    def db_engine(self) -> Engine:
        return self._db_engine

    # Database Management API

    def create_database(self, name: str, skip_exists: bool = False):
        return create_database(self._db_engine, name, skip_exists)

    def drop_database(self, name: str):
        db_name = self._identifier_preparer.quote(name)
        with self._db_engine.connect() as conn:
            stmt = text(f"DROP DATABASE IF EXISTS {db_name};")
            return conn.execute(stmt)

    def database_names(self) -> List[str]:
        stmt = text("SHOW DATABASES;")
        with self._db_engine.connect() as conn:
            result = conn.execute(stmt)
            return [row[0] for row in result]

    def has_database(self, name: str) -> bool:
        return database_exists(self._db_engine, name)

    # Table Management API

    def create_table(
        self,
        *,
        schema: Optional[Type[TableModel]] = None,
        mode: Optional[Literal["create", "overwrite", "exist_ok"]] = "create",
    ) -> Table:
        if mode == "create":
            table = Table(schema=schema, client=self)
        elif mode == "overwrite":
            self.drop_table(schema.__tablename__)
            table = Table(schema=schema, client=self)
        elif mode == "exist_ok":
            table = Table(schema=schema, client=self, exist_ok=True)
        else:
            raise ValueError(f"Invalid create mode: {mode}")
        return table

    def _get_table_model(self, table_name: str) -> Optional[Type[DeclarativeMeta]]:
        for m in default_registry.mappers:
            if m.persist_selectable.name == table_name:
                return m.class_
        return None

    def open_table(self, table_name: str) -> Optional[Table]:
        # If the table in the mapper registry.
        table_model = self._get_table_model(table_name)
        if table_model is not None:
            table = Table(schema=table_model, client=self, exist_ok=True)
            return table

        return None

    def table_names(self) -> List[str]:
        return self._inspector.get_table_names()

    def has_table(self, table_name: str) -> bool:
        return self._inspector.has_table(table_name)

    def drop_table(self, table_name: str):
        table = sqlalchemy.Table(
            table_name, Base.metadata, autoload_with=self._db_engine
        )
        return table.drop(self._db_engine)

    # Raw SQL API

    def execute(
        self,
        sql: str | Executable,
        params: Optional[dict] = None,
        raise_error: Optional[bool] = False,
    ) -> SQLExecuteResult:
        try:
            with self.session() as session:
                if isinstance(sql, str):
                    stmt = text(sql)
                else:
                    stmt = sql
                result: Result = session.execute(stmt, params or {})
                return SQLExecuteResult(rowcount=result.rowcount, success=True)
        except Exception as e:
            if raise_error:
                raise e
            logger.error(f"Failed to execute SQL: {str(e)}")
            return SQLExecuteResult(rowcount=0, success=False, message=str(e))

    def query(
        self,
        sql: str | SelectBase,
        params: Optional[dict] = None,
    ) -> SQLQueryResult:
        with self.session() as session:
            if isinstance(sql, str):
                stmt = text(sql)
            else:
                stmt = sql
            result = session.execute(stmt, params)
            return SQLQueryResult(result)

    # Session API

    @contextmanager
    def session(
        self, *, provided_session: Optional[Session] = None, **kwargs
    ) -> Generator[Session, None, None]:
        if provided_session is not None:
            session = provided_session
            is_local_session = False
        elif SESSION.get() is not None:
            session = SESSION.get()
            is_local_session = False
        else:
            # Since both the TiDB Client and Table API begin a Session within the method, the Session ends when
            # the method returns. The error: "Parent instance <x> is not bound to a Session;" will show when accessing
            # the returned object. To prevent it, we set the expire_on_commit parameter to False by default.
            # Details: https://sqlalche.me/e/20/bhk3
            kwargs.setdefault("expire_on_commit", False)
            session = Session(self._db_engine, **kwargs)
            SESSION.set(session)
            is_local_session = True

        try:
            yield session
            if is_local_session:
                session.commit()
        except Exception as e:
            session.rollback()
            raise e
        finally:
            if is_local_session:
                session.close()
                SESSION.set(None)



from typing import (
    Literal,
    Optional,
    List,
    Any,
    Dict,
    TypeVar,
    Type,
    Union,
    TYPE_CHECKING,
)
import warnings

from sqlalchemy import Engine, Table as SaTable
from sqlalchemy.orm import DeclarativeMeta
from sqlmodel import Session
from sqlmodel.main import SQLModelMetaclass
from typing_extensions import Generic

from pytidb.base import Base
from pytidb.filters import Filters, build_filter_clauses
from pytidb.orm.indexes import FullTextIndex, VectorIndex, format_distance_expression
from pytidb.sql import select, update, delete
from pytidb.schema import (
    QueryBundle,
    TableModelMeta,
    VectorDataType,
    TableModel,
    ColumnInfo,
)
from pytidb.search import SearchType, SearchQuery
from pytidb.result import QueryResult, SQLModelQueryResult
from pytidb.utils import (
    check_text_column,
    check_vector_column,
    filter_text_columns,
    filter_vector_columns,
    get_index_type,
)

if TYPE_CHECKING:
    from pytidb import TiDBClient


T = TypeVar("T", bound=TableModel)


class Table(Generic[T]):
    def __init__(
        self,
        *,
        client: "TiDBClient",
        schema: Optional[Type[T]] = None,
        exist_ok: bool = False,
    ):
        self._client = client
        self._db_engine = client.db_engine
        self._identifier_preparer = self._db_engine.dialect.identifier_preparer

        # Init table model.
        if (
            type(schema) is TableModelMeta
            or type(schema) is SQLModelMetaclass
            or type(schema) is DeclarativeMeta
        ):
            self._table_model = schema
        else:
            raise TypeError(f"Invalid schema type: {type(schema)}")

        self._sa_table: SaTable = self._table_model.__table__
        self._columns = self._table_model.__table__.columns

        # Determine default vector column for vector search.
        self._default_vector_column = None
        self._vector_columns = filter_vector_columns(self._columns)
        if len(self._vector_columns) == 1:
            self._default_vector_column = self._vector_columns[0]

        # Determine default text column for fulltext search.
        self._default_text_column = None
        self._text_columns = filter_text_columns(self._columns)
        if len(self._text_columns) == 1:
            self._default_text_column = self._text_columns[0]

        # Setup auto embedding.
        if hasattr(schema, "__pydantic_fields__"):
            vector_fields = {}
            text_fields = {}

            for field_name, field_info in self._table_model.__pydantic_fields__.items():
                if field_info._attributes_set.get("field_type", None) == "vector":
                    vector_fields[field_name] = field_info
                elif field_info._attributes_set.get("field_type", None) == "text":
                    text_fields[field_name] = field_info

            if len(vector_fields) == 1 and self._default_vector_column is None:
                vector_field_name = list(vector_fields.keys())[0]
                self._default_vector_column = check_vector_column(
                    self._columns, vector_field_name
                )
            if len(text_fields) == 1 and self._default_text_column is None:
                text_field_name = list(text_fields.keys())[0]
                self._default_text_column = check_text_column(
                    self._columns, text_field_name
                )

            self._setup_auto_embedding(vector_fields)
            self._auto_create_vector_index(vector_fields)
            self._auto_create_fulltext_index(text_fields)

        # Create table.
        Base.metadata.create_all(
            self._db_engine, tables=[self._sa_table], checkfirst=exist_ok
        )

    @property
    def table_model(self) -> T:
        return self._table_model

    @property
    def table_name(self) -> str:
        return self._table_model.__tablename__

    @property
    def client(self) -> "TiDBClient":
        return self._client

    @property
    def db_engine(self) -> Engine:
        return self._db_engine

    @property
    def vector_columns(self):
        return self._vector_columns

    @property
    def text_columns(self):
        return self._text_columns

    @property
    def auto_embedding_configs(self):
        return self._auto_embedding_configs

    def _setup_auto_embedding(self, vector_fields):
        """Setup auto embedding configurations for fields with embed_fn."""
        self._auto_embedding_configs = {}
        for vector_field_name, field in vector_fields.items():
            embed_fn = field._attributes_set.get("embed_fn", None)
            if embed_fn is None:
                continue

            source_field_name = field._attributes_set.get("source_field", None)
            if source_field_name is None:
                continue

            self._auto_embedding_configs[vector_field_name] = {
                "embed_fn": embed_fn,
                "vector_field": field,
                "vector_field_name": vector_field_name,
                "source_field_name": source_field_name,
            }

    def _auto_create_vector_index(self, vector_fields):
        for field_name, field in vector_fields.items():
            column_name = field_name
            skip_index = field._attributes_set.get("skip_index", False)
            if skip_index:
                continue

            distance_metric = field._attributes_set.get("distance_metric", "COSINE")
            algorithm = field._attributes_set.get("algorithm", "HNSW")

            # Check if the metric on the column is already defined in vector indexes
            distance_expression = format_distance_expression(
                column_name, distance_metric
            )
            indexed_expressions = [
                index.expressions[0].text
                for index in self._sa_table.indexes
                if get_index_type(index) == "vector"
            ]
            if distance_expression in indexed_expressions:
                continue

            # Create vector index automatically, if not defined
            self._sa_table.append_constraint(
                VectorIndex(
                    f"vec_idx_{column_name}_{distance_metric.lower()}",
                    column_name,
                    distance_metric=distance_metric,
                    algorithm=algorithm,
                )
            )

    def _auto_create_fulltext_index(self, text_fields):
        for text_field_name, field in text_fields.items():
            skip_index = field._attributes_set.get("skip_index", False)
            if skip_index:
                continue

            # Check if the column is already defined a fulltext index.
            column_name = text_field_name
            indexed_columns = [
                index.columns[0].name
                for index in self._sa_table.indexes
                if get_index_type(index) == "fulltext"
            ]
            if column_name in indexed_columns:
                continue

            # Create fulltext index automatically, if not defined
            fts_parser = field._attributes_set.get("fts_parser", "MULTILINGUAL")
            self._sa_table.append_constraint(
                FullTextIndex(
                    f"fts_idx_{column_name}",
                    column_name,
                    fts_parser=fts_parser,
                )
            )

    def get(self, id: Any) -> T:
        with self._client.session() as db_session:
            return db_session.get(self._table_model, id)

    def insert(self, data: T) -> T:
        # Auto embedding.
        for field_name, config in self._auto_embedding_configs.items():
            if getattr(data, field_name) is not None:
                # Vector embeddings is provided.
                continue

            if not hasattr(data, config["source_field_name"]):
                continue

            embedding_source = getattr(data, config["source_field_name"])
            vector_embedding = config["embed_fn"].get_source_embedding(embedding_source)
            setattr(data, field_name, vector_embedding)

        with self._client.session() as db_session:
            db_session.add(data)
            db_session.flush()
            db_session.refresh(data)
            return data

    def bulk_insert(self, data: List[T]) -> List[T]:
        # Auto embedding.
        for field_name, config in self._auto_embedding_configs.items():
            items_need_embedding = []
            sources_to_embedding = []

            # Skip if no embedding function is provided.
            if "embed_fn" not in config or config["embed_fn"] is None:
                continue

            for item in data:
                # Skip if vector embeddings is provided.
                if getattr(item, field_name) is not None:
                    continue

                # Skip if no source field is provided.
                if not hasattr(item, config["source_field_name"]):
                    continue

                items_need_embedding.append(item)
                embedding_source = getattr(item, config["source_field_name"])
                sources_to_embedding.append(embedding_source)

            # Batch embedding.
            vector_embeddings = config["embed_fn"].get_source_embeddings(
                sources_to_embedding
            )
            for item, embedding in zip(items_need_embedding, vector_embeddings):
                setattr(item, field_name, embedding)

        with self._client.session() as db_session:
            db_session.add_all(data)
            db_session.flush()
            for item in data:
                db_session.refresh(item)
            return data

    def update(self, values: dict, filters: Optional[Filters] = None) -> object:
        # Auto embedding.
        for field_name, config in self._auto_embedding_configs.items():
            if field_name in values:
                # Vector embeddings is provided.
                continue

            if config["source_field_name"] not in values:
                continue

            embedding_source = values[config["source_field_name"]]
            vector_embedding = config["embed_fn"].get_source_embedding(embedding_source)
            values[field_name] = vector_embedding

        with self._client.session() as db_session:
            filter_clauses = build_filter_clauses(
                filters, self._columns, self._table_model
            )
            stmt = update(self._table_model).filter(*filter_clauses).values(values)
            db_session.execute(stmt)

    def delete(self, filters: Optional[Filters] = None):
        """
        Delete data from the TiDB table.

        params:
            filters: (Optional[Dict[str, Any]]): The filters to apply to the delete operation.
        """
        with self._client.session() as db_session:
            filter_clauses = build_filter_clauses(
                filters, self._columns, self._table_model
            )
            stmt = delete(self._table_model).filter(*filter_clauses)
            db_session.execute(stmt)

    def truncate(self):
        with self._client.session():
            table_name = self._identifier_preparer.quote(self.table_name)
            stmt = f"TRUNCATE TABLE {table_name};"
            self._client.execute(stmt)

    def columns(self) -> List[ColumnInfo]:
        with self._client.session():
            table_name = self._identifier_preparer.quote(self.table_name)
            stmt = """
                SELECT
                    COLUMN_NAME as column_name,
                    COLUMN_TYPE as column_type
                FROM INFORMATION_SCHEMA.COLUMNS
                WHERE
                    TABLE_SCHEMA = DATABASE()
                    AND TABLE_NAME = :table_name;
            """
            res = self._client.query(stmt, {"table_name": table_name})
            return res.to_pydantic(ColumnInfo)

    def rows(self):
        with self._client.session():
            table_name = self._identifier_preparer.quote(self.table_name)
            stmt = f"SELECT COUNT(*) FROM {table_name};"
            return self._client.query(stmt).scalar()

    def query(
        self,
        filters: Optional[Filters] = None,
        order_by: Optional[List[Any] | str | Dict[str, Any]] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> QueryResult:
        with Session(self._db_engine) as db_session:
            stmt = select(self._table_model)

            # Apply filters.
            if filters is not None:
                filter_clauses = build_filter_clauses(
                    filters, self._columns, self._table_model
                )
                stmt = stmt.filter(*filter_clauses)

            # Apply order by.
            if isinstance(order_by, list):
                stmt = stmt.order_by(*order_by)
            elif isinstance(order_by, str):
                if order_by not in self._columns:
                    raise KeyError(f"Unknown order by column: {order_by}")
                stmt = stmt.order_by(self._columns[order_by])
            elif isinstance(order_by, dict):
                for key, value in order_by.items():
                    if key not in self._columns:
                        raise KeyError(f"Unknown order by column: {key}")

                    if value == "desc":
                        stmt = stmt.order_by(self._columns[key].desc())
                    elif value == "asc":
                        stmt = stmt.order_by(self._columns[key])
                    else:
                        raise ValueError(
                            f"Invalid order direction value (allowed: 'desc', 'asc'): {value}"
                        )

            # Pagination.
            if limit is not None:
                stmt = stmt.limit(limit)
            if offset is not None:
                stmt = stmt.offset(offset)

            result = db_session.exec(stmt).all()
            return SQLModelQueryResult(result)

    def search(
        self,
        query: Optional[Union[VectorDataType, str, QueryBundle]] = None,
        search_type: SearchType = "vector",
    ) -> SearchQuery:
        return SearchQuery(
            table=self,
            query=query,
            search_type=search_type,
        )

    def _has_tiflash_index(
        self,
        column_name: str,
        index_kind: Optional[Literal["FullText", "Vector"]] = None,
    ) -> bool:
        stmt = """SELECT EXISTS(
            SELECT 1
            FROM INFORMATION_SCHEMA.TIFLASH_INDEXES
            WHERE
                TIDB_DATABASE = DATABASE()
                AND TIDB_TABLE = :table_name
                AND COLUMN_NAME = :column_name
                AND INDEX_KIND = :index_kind
        )
        """
        with self._client.session():
            res = self._client.query(
                stmt,
                {
                    "table_name": self.table_name,
                    "column_name": column_name,
                    "index_kind": index_kind,
                },
            )
            return res.scalar()

    def has_vector_index(self, column_name: str) -> bool:
        return self._has_tiflash_index(column_name, "Vector")

    def has_fts_index(self, column_name: str) -> bool:
        return self._has_tiflash_index(column_name, "FullText")

    def create_vector_index(self, column_name: str, name: Optional[str] = None):
        # TODO: Support exist_ok.
        warnings.warn(
            "table.create_vector_index() is an experimental API, use VectorField instead."
        )
        index_name = name or f"vec_idx_{column_name}"
        vec_idx = VectorIndex(index_name, self._columns[column_name])
        vec_idx.create(self.client.db_engine)

    def create_fts_index(
        self, column_name: str, name: Optional[str] = None, exist_ok: bool = False
    ):
        warnings.warn(
            "table.create_fts_index() is an experimental API, use FullTextField instead."
        )
        index_name = name or f"fts_idx_{column_name}"
        fts_idx = FullTextIndex(index_name, self._columns[column_name])
        fts_idx.create(self.client.db_engine, checkfirst=exist_ok)



from sqlalchemy import Engine, text


def create_database(db_engine: Engine, name: str, skip_exists: bool = False):
    identifier_preparer = db_engine.dialect.identifier_preparer
    db_name = identifier_preparer.quote(name)
    with db_engine.connect() as conn:
        if skip_exists:
            stmt = text(f"CREATE DATABASE IF NOT EXISTS {db_name};")
        else:
            stmt = text(f"CREATE DATABASE {db_name};")
        return conn.execute(stmt)


def database_exists(db_engine: Engine, name: str) -> bool:
    stmt = text(
        "SELECT SCHEMA_NAME FROM INFORMATION_SCHEMA.SCHEMATA WHERE SCHEMA_NAME = :db_name;"
    )
    with db_engine.connect() as conn:
        result = conn.execute(stmt, {"db_name": name})
        return bool(result.scalar())



from typing import Literal
from sqlalchemy import text
from sqlalchemy.sql.schema import Index
from pytidb.orm.sql.ddl import TiDBSchemaGenerator

"""
The algorithm used for vector index.

Available algorithms:
  HNSW: Hierarchical Navigable Small World graph.
"""
VectorIndexAlgorithm = Literal["HNSW"]


"""
The distance metric can be used for vector index.

Available distance metrics:
  COSINE: Cosine distance metric.
  L2: L2 (Euclidean) distance metric.
"""
DistanceMetric = Literal["COSINE", "L2"]

distance_fn_mappings: dict[DistanceMetric, str] = {
    "COSINE": "VEC_COSINE_DISTANCE",
    "L2": "VEC_L2_DISTANCE",
}


def format_distance_expression(
    column_name: str, distance_metric: DistanceMetric
) -> str:
    distance_fn = distance_fn_mappings[distance_metric]
    return f"({distance_fn}({column_name}))"


class VectorIndex(Index):
    """Vector Index schema."""

    def __init__(
        self,
        name,
        *columns,
        distance_metric: DistanceMetric = "COSINE",
        algorithm: VectorIndexAlgorithm = "HNSW",
        ensure_columnar_replica: bool = True,
        **kw,
    ):
        if len(columns) == 0:
            raise ValueError("Vector index must have at least one column")
        if len(columns) > 1:
            raise ValueError(
                f"Vector index can only apply to one column, but got {len(columns)} columns"
            )
        if algorithm not in ["HNSW"]:
            raise ValueError(f"Invalid vector index algorithm: {algorithm}")
        if distance_metric not in ["COSINE", "L2"]:
            raise ValueError(f"Invalid distance metric: {distance_metric}")

        # Convert column name to distance expression.
        vector_column = columns[0]
        table = kw.get("table", None)
        if isinstance(vector_column, str):
            vector_column_name = vector_column
            distance_expression = format_distance_expression(
                vector_column_name, distance_metric
            )
        elif hasattr(vector_column, "table"):
            table = vector_column.table
            vector_column_name = vector_column.name
            distance_expression = format_distance_expression(
                vector_column_name, distance_metric
            )
        else:
            raise ValueError(f"Invalid vector column: {vector_column}")
        expressions = [text(distance_expression)]

        self.distance_metric = distance_metric
        self.ensure_columnar_replica = ensure_columnar_replica
        kw["mysql_prefix"] = "VECTOR"
        kw["mysql_using"] = algorithm
        super().__init__(name, *expressions, _table=table, **kw)

    def create(self, bind, checkfirst: bool = False) -> None:
        bind._run_ddl_visitor(TiDBSchemaGenerator, self, checkfirst=checkfirst)


"""
Parser used for full text index.

Available parsers:
  STANDARD: Fast parser for English content, splits words by spaces and punctuation.
  MULTILINGUAL: Supports multiple languages including English, Chinese, Japanese, and Korean.
"""
FullTextParser = Literal["STANDARD", "MULTILINGUAL"]


class FullTextIndex(Index):
    """Full Text Index schema."""

    def __init__(
        self,
        name,
        *column_names,
        fts_parser: FullTextParser = "MULTILINGUAL",
        ensure_columnar_replica: bool = True,
        **kw,
    ):
        if fts_parser not in ["STANDARD", "MULTILINGUAL"]:
            raise ValueError(f"Invalid full text parser: {fts_parser}")
        self.ensure_columnar_replica = ensure_columnar_replica
        kw["mysql_prefix"] = "FULLTEXT"
        kw["mysql_with_parser"] = fts_parser
        super().__init__(name, *column_names, **kw)
